#!/usr/bin/env python
# -*- mode: python; -*-

"""
AIA Tools provides a set of Python classes to interface with application sources exported from
`App Inventor <http://appinventor.mit.edu/>`_.
"""

from aiatools.aia import *
from aiatools.algebra import *
from aiatools.attributes import *
from aiatools.block_types import *
from aiatools.component_types import *
from aiatools.selectors import *

__author__ = 'Evan W. Patton <ewpatton@mit.edu>'
__version__ = '0.4.1'


def _aia_main():
    """
    The main program for aiatools. See aiatool --help for more information.
    """
    raise NotImplemented()


if __name__ == '__main__':
    _aia_main()
